import pprint

import typer
from halo import Halo
from tabulate import tabulate

import preserve


app = typer.Typer()

# TODO better implementation of multi
# TODO support with statements in preserve.open
# TODO documentation
# TODO tests


@app.callback()
def callback():
    """🥫 Preserve - A simple Key/Value database with multiple backends."""


@app.command(help="Export a database to a different database.")
def export(
    input: str = typer.Option(
        ...,
        "--input",
        "-i",
        help="The URI specifying how to access the input preserve database.",
    ),
    output: str = typer.Option(
        ...,
        "--output",
        "-o",
        help="The URI specifying how to access the input preserve database.",
    ),
):
    assert input != output

    with Halo(text=f"Opening input preserve: {input}.", spinner="dots") as sp:
        in_preserve_db = preserve.from_uri(input)
        sp.succeed()

    with Halo(
        text=f"Opening output preserve: {output}.", spinner="dots"
    ) as sp:
        out_preserve_db = preserve.from_uri(output)
        sp.succeed()

    with typer.progressbar(
        list(in_preserve_db), label="Exporting data"
    ) as progress:
        for idx in progress:
            out_preserve_db[idx] = in_preserve_db[idx]

    in_preserve_db.close()
    out_preserve_db.close()


@app.command(help="List available connectors.")
def connectors(
    # connector: Optional[str] = typer.Option(
    #    None, "--connector", "-c", help="The connector to inspect."
    # )
):

    values = []
    for c in preserve.connectors():
        sc = c.schema()
        props_values = []
        for n, p in sc["properties"].items():
            props = [n, f"{p['type'].upper()}"]
            if "default" in p:
                props.append(f"[default: {p['default']}]")

            props_values.append(props)

        values.append(
            [
                c.scheme(),
                sc["description"],
                tabulate(props_values, showindex="never", tablefmt="plain"),
            ]
        )

    typer.echo(
        tabulate(
            values,
            headers=["scheme", "description", "parameters"],
            showindex="never",
            tablefmt="psql",
        )
    )


@app.command(help="Get header of a given database table.")
def header(
    uri: str = typer.Argument(
        ..., help="The URI specifying how to access the preserve database."
    ),
    rows: int = typer.Option(
        10, "--nb", "-n", help="The number of rows to display."
    ),
):
    pp = pprint.PrettyPrinter(width=41, compact=True)

    preserve_db = None
    with Halo(text=f"Opening preserve: {uri}.", spinner="dots") as sp:
        preserve_db = preserve.from_uri(uri)
        sp.succeed()

    if preserve_db:

        with Halo(text="Fetching head.", spinner="dots") as sp:
            values = [
                [idx, pp.pformat(preserve_db[idx])]
                for idx in list(preserve_db)[: min(len(preserve_db), rows)]
            ]
            sp.succeed()

            typer.echo(
                tabulate(
                    values,
                    headers=["key", "value"],
                    showindex="never",
                    tablefmt="psql",
                )
            )

            preserve_db.close()


if __name__ == "__main__":
    app()



import collections.abc
from abc import abstractmethod

from pydantic import BaseModel


# TODO Add typing
class Connector(BaseModel, collections.abc.MutableMapping):
    """Similar to Shelve code."""

    @staticmethod
    @abstractmethod
    def scheme() -> str:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_uri(uri: str) -> "Connector":
        raise NotImplementedError()

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    @abstractmethod
    def __len__(self):
        raise NotImplementedError()

    @abstractmethod
    def __contains__(self, key):
        raise NotImplementedError()

    @abstractmethod
    def get(self, key, default=None):
        raise NotImplementedError()

    @abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError()

    @abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError()

    @abstractmethod
    def __delitem__(self, key):
        raise NotImplementedError()

    @abstractmethod
    def __enter__(self):
        raise NotImplementedError()

    @abstractmethod
    def __exit__(self, type, value, traceback):
        raise NotImplementedError()

    @abstractmethod
    def close(self):
        raise NotImplementedError()

    @abstractmethod
    def __del__(self):
        raise NotImplementedError()

    @abstractmethod
    def sync(self):
        raise NotImplementedError()

    def __setattr__(self, attr, value):
        if attr in self.__slots__:
            object.__setattr__(self, attr, value)
        else:
            super().__setattr__(attr, value)



# -*- coding: utf-8 -*-

from preserve.connector import Connector
from preserve.preserve import Preserve, connectors, from_uri, open


__all__ = ["Connector", "Preserve", "connectors", "from_uri", "open"]


def __get_pyproject():
    import io
    import os

    import toml

    init_path = os.path.abspath(os.path.dirname(__file__))
    pyproject_path = os.path.join(init_path, "../pyproject.toml")

    with io.open(pyproject_path, "r") as fopen:
        pyproject = toml.load(fopen)

    return pyproject["tool"]["poetry"]


__author__ = "Grégoire Burel"
__copyright__ = "Grégoire Burel"
__license__ = "MIT"

__version__ = __get_pyproject()["version"]
__doc__ = __get_pyproject()["description"]



import shelve
from typing import Optional, cast
from urllib import parse

from preserve.connector import Connector


class Shelf(Connector):
    """Preserve connector using Shelf backend."""

    filename: str
    protocol: Optional[str] = None
    writeback: bool = False
    keyencoding: str = "utf-8"

    __slots__ = ["_shelf"]

    def __setattr__(self, attr, value):
        if attr in self.__slots__:
            object.__setattr__(self, attr, value)
        else:
            super().__setattr__(attr, value)

    @staticmethod
    def scheme() -> str:
        return "shelf"

    # example: shelf://filename?protocol=?,writeback=?, keyencoding="utf-8"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._shelf = shelve.DbfilenameShelf(
            self.filename,
            flag="c",
            protocol=self.protocol,
            writeback=self.writeback,
        )

    @classmethod
    def from_uri(cls, uri: str) -> "Shelf":
        p = parse.urlsplit(uri)
        if p.scheme != cls.scheme():
            raise ValueError()

        params = {}
        params["filename"] = p.path
        params.update(dict(parse.parse_qsl(p.query)))

        return cast("Shelf", cls.parse_obj(params))

    def __iter__(self):
        return self._shelf.__iter__()

    def __len__(self):
        return self._shelf.__len__()

    def __contains__(self, key):
        return self._shelf.__contains__(key)

    def get(self, key, default=None):
        return self._shelf.get(key, default=default)

    def __getitem__(self, key):
        return self._shelf.__getitem__(key)

    def __setitem__(self, key, value):
        self._shelf.__setitem__(key, value)

    def __delitem__(self, key):
        self._shelf.__delitem__(key)

    def __enter__(self):
        return self._shelf

    def __exit__(self, type, value, traceback):
        self._shelf.close()

    def close(self):
        self._shelf.close()

    def __del__(self):
        self.close()

    def sync(self):
        self._shelf.sync()



from preserve.connectors.memory import Memory
from preserve.connectors.mongo import Mongo  # , MultiMongo
from preserve.connectors.shelf import Shelf  # , MultiShelf


__all__ = ["Memory", "Mongo", "Shelf"]



import importlib
import inspect
import logging
import pkgutil
import sys
from abc import abstractmethod
from typing import Dict, List, Type

import pkg_resources
from pydantic import BaseModel

import preserve.connectors as _connectors
from preserve.connector import Connector

from .utils import Singleton


_logger = logging.getLogger(__name__)


def _iter_namespace(ns_pkg):
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


class Preserve(object, metaclass=Singleton):
    # import preserve.jars

    def __init__(self):
        self._connectors: Dict[str, Type[Connector]] = {}

        # Discover new plugins
        # https://packaging.python.org/guides/creating-and-discovering-plugins/:
        # TODO Check if it actually works
        for entry_point in pkg_resources.iter_entry_points(
            "preserve.connectors"
        ):
            print(entry_point.load())
            self._connectors[entry_point.name] = entry_point.load()

        # Register all the connectors that are defined in the package:
        for finder, name, ispkg in _iter_namespace(_connectors):
            module = importlib.import_module(name)
            for _, cls in inspect.getmembers(
                sys.modules[module.__name__],
                predicate=lambda o: inspect.isclass(o)
                and issubclass(o, Connector)
                and o != Connector,
            ):
                self._connectors[cls.scheme()] = cls

        # print(self._connectors)

    def open(self, format: str, **kwargs) -> Connector:
        if format not in self._connectors:
            raise ValueError(format)

        return self._connectors[format](**kwargs)

    def from_uri(self, uri: str) -> Connector:
        if ":" not in uri:
            raise ValueError(uri)

        f = uri.split(":", 1)[0]
        if f not in self._connectors:
            raise ValueError(f)

        return self._connectors[f].from_uri(uri)

    def register(self, format: str, connector):
        self._connectors[format] = connector

    def is_registerd(self, format: str, connector: Type[Connector]) -> bool:
        if (
            format in self._connectors
            and self._connectors[format] == connector
        ):
            return True
        return False

    def connectors(self) -> List[Type[Connector]]:
        return list(self._connectors.values())


def open(format: str, **kwargs) -> Connector:
    return Preserve().open(format, **kwargs)


def from_uri(uri: str) -> Connector:
    return Preserve().from_uri(uri)


def connectors() -> List[Type[Connector]]:
    return Preserve().connectors()



from typing import Any, Dict, Optional, cast
from urllib import parse

from pymongo import MongoClient

from preserve.connector import Connector


class Mongo(Connector):
    """Mongodb backend for preserve."""

    database: str = "db"
    collection: Optional[str] = None
    host: str = "127.0.0.1"
    port: int = 27017

    __slots__ = ["_client", "_collection"]

    @classmethod
    def from_uri(cls, uri: str) -> "Mongo":
        p = parse.urlsplit(uri)
        if p.scheme != cls.scheme():
            raise ValueError()

        params: Dict[str, Any] = {}
        if p.hostname:
            params["host"] = p.hostname

        if p.port:
            params["port"] = p.port

        if p.path and len(p.path.split("/", 1)) > 1:
            params["database"] = p.path.split("/", 1)[1]

        params.update(dict(parse.parse_qsl(p.query)))

        return cast("Mongo", cls.parse_obj(params))

    @staticmethod
    def scheme() -> str:
        return "mongodb"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.collection:
            self.collection = self.database

        self._client = MongoClient(self.host, self.port)
        self._collection = self._client[self.database][self.collection]

    def __iter__(self):
        for item in self._collection.find():
            yield item["_id"]

    def items(self):
        for item in self._collection.find():
            _id = item["_id"]
            del item["_id"]
            yield _id, item

    def __len__(self):
        return self._collection.count_documents({})

    def __contains__(self, key):
        if self._collection.count_documents({"_id": key}, limit=1) != 0:
            return True
        else:
            return False

    def get(self, key, default=None):
        item = self.__getitem__(key)
        return item if item else default

    def __getitem__(self, key):
        item = self._collection.find_one({"_id": key})
        if item:
            del item["_id"]
        return item

    def __setitem__(self, key, value):
        item = dict(value)
        item["_id"] = key
        self._collection.update_one({"_id": key}, {"$set": item}, upsert=True)

    def __delitem__(self, key):
        self._collection.delete_one({"_id": key})

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._client.close()

    def close(self):
        self._client.close()

    def __del__(self):
        self.close()

    def sync(self):
        pass



# from urllib.parse import urlparse
from typing import Any


class Singleton(type):
    _instances: Any = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs
            )
        return cls._instances[cls]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['preserve', 'preserve.connectors']

package_data = \
{'': ['*']}

install_requires = \
['halo>=0.0.30,<0.0.31',
 'pydantic>=1.6.1,<2.0.0',
 'pymongo>=3.11.0,<4.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'toml>=0.10.1,<0.11.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['preserve = preserve.cli:app']}

setup_kwargs = {
    'name': 'preserve',
    'version': '0.1.2a0',
    'description': 'A simple key/value store with multiple backends.',
    'long_description': '# 🥫 Preserve - A simple Python Key/Value database with multiple backends.\n\n> ⚠️ Preserve is alpha software and currently in development (i.e., no tests).\n\nPreserve is a simple (simplistic) key/value store for storing JSON-like data in different backends. Its API is based on the standard Python dictionary API.\n\n\n## ℹ️ Installation and Usage\nPreserve can be installed using pip:\n```\npip install preserve\n```\n\nPreserve can be also installed from Github directly using the following command:\n```\npip install git+https://github.com/evhart/preserve#egg=preserve\n```\n\n### 📒 Requirements\nPreserve needs the following libraries installed and Python 3.6+ (tested on Python 3.8):\n* [halo](https://github.com/manrajgrover/halo)\n* [pydantic](https://pydantic-docs.helpmanual.io/)\n* [pymongo](https://pymongo.readthedocs.io/)\n* [tabulate](https://github.com/astanin/python-tabulate)\n* [typer](https://typer.tiangolo.com/)\n\n### 🐍 Python API\n\nIf you know how to use Python dictionaries, you already know how to use preserve. Simply use the backend connector that corresponds to your database and you are ready to go.\n\nYou can either create a new database from a standarised database URI or using the driver parameters:\n\n```python\nimport preserve\n\n# Using parameters:\njam_db1 = preserve.open(\'shelf\', filename="preserve.dbm")\njam_db1[\'strawberry\'] = {\'name\': \'Strawbery Jam\', \'ingredients\': [\'strawberry\', \'sugar\']}\n\n\n# Using URI:\njam_db2 = preserve.from_uri("mongodb://127.0.0.1:27017/preserves?collection=jam")\njam_db2[\'currant\'] = {\'name\': \'Currant Jam\', \'ingredients\': [\'currant\', \'sugar\']}\n\n```\n\n### 🖥️ Command Line Interface (CLI)\nPreserve has a simple CLI utility that can be access using the ```preserve``` command. Preserve support migrating/exporting data from one database ot another database and showing the firs rows from databases.\n\n```\nUsage: preserve [OPTIONS] COMMAND [ARGS]...\n\n  🥫 Preserve - A simple Key/Value database with multiple backends.\n\nOptions:\n  --install-completion [bash|zsh|fish|powershell|pwsh]\n                                  Install completion for the specified shell.\n  --show-completion [bash|zsh|fish|powershell|pwsh]\n                                  Show completion for the specified shell, to\n                                  copy it or customize the installation.\n\n  --help                          Show this message and exit.\n\nCommands:\n  connectors  List available connectors.\n  export      Export a database to a different output.\n  header      Get header of a given database table.\n```\n',
    'author': 'Grégoire Burel',
    'author_email': 'evhart@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/evhart/preserve/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
